#ifndef SCINTDIGIALGS_SCINTWAVEFORMDIGIALG_H
#define SCINTDIGIALGS_SCINTWAVEFORMDIGIALG_H

// Base class
#include "AthenaBaseComps/AthReentrantAlgorithm.h"

// Data classes
#include "WaveRawEvent/RawWaveformContainer.h"
#include "ScintSimEvent/ScintHitCollection.h"

// Tool classes
#include "WaveDigiTools/IWaveformDigitisationTool.h"
#include "WaveformConditionsTools/IWaveformCableMappingTool.h"
#include "WaveformConditionsTools/IWaveformDigiConditionsTool.h"

// Handles
#include "StoreGate/ReadHandleKey.h"
#include "StoreGate/WriteHandleKey.h"

// Gaudi
#include "GaudiKernel/ServiceHandle.h"
#include "GaudiKernel/ToolHandle.h"

//Helpers
#include "ScintIdentifier/VetoID.h"
#include "ScintIdentifier/VetoNuID.h"
#include "ScintIdentifier/TriggerID.h"
#include "ScintIdentifier/PreshowerID.h"
#include "ScintSimEvent/ScintHitIdHelper.h"
#include "Identifier/Identifier.h"


// ROOT
#include "TF1.h"


// STL
#include <string>
#include <vector>

class ScintWaveformDigiAlg : public AthReentrantAlgorithm {

 public:
  // Constructor
  ScintWaveformDigiAlg(const std::string& name, ISvcLocator* pSvcLocator);
  virtual ~ScintWaveformDigiAlg() = default;

  /** @name Usual algorithm methods */
  //@{
  virtual StatusCode initialize() override;
  virtual StatusCode execute(const EventContext& ctx) const override;
  virtual StatusCode finalize() override;
  //@}


 private:

  /** @name Disallow default instantiation, copy, assignment */
  //@{
  ScintWaveformDigiAlg() = delete;
  ScintWaveformDigiAlg(const ScintWaveformDigiAlg&) = delete;
  ScintWaveformDigiAlg &operator=(const ScintWaveformDigiAlg&) = delete;
  //@}

  /// 

  // We use these a lot, so read them once from the conditions tool
  mutable float m_base_mean;
  mutable float m_base_rms;


  /** Kernel PDF and evaluated values **/
  //@{
  mutable TF1*                            m_kernel;
  mutable std::vector<float>              m_timekernel;
  //@}
  

  /// Detector ID helpers
  const VetoID* m_vetoID{nullptr};
  const VetoNuID* m_vetoNuID{nullptr};
  const TriggerID* m_triggerID{nullptr};
  const PreshowerID* m_preshowerID{nullptr};

  /**
   * @name Digitisation tool
   */
  ToolHandle<IWaveformDigitisationTool> m_digiTool
    {this, "WaveformDigitisationTool", "WaveformDigitisationTool"};

  /**
   * @name Mapping tool
   */
  ToolHandle<IWaveformCableMappingTool> m_mappingTool
    {this, "WaveformCableMappingTool", "WaveformCableMappingTool"};

  /**
   * @name Digitization parameters tool
   */
  ToolHandle<IWaveformDigiConditionsTool> m_digiCondTool
    {this, "DigiConditionsTool", ""};

  /**
   * @name Input HITS using SG::ReadHandleKey
   */
  //@{

  SG::ReadHandleKey<ScintHitCollection> m_scintHitContainerKey 
  {this, "ScintHitContainerKey", ""};

  //@}


  /**
   * @name Output data using SG::WriteHandleKey
   */
  //@{
  SG::WriteHandleKey<RawWaveformContainer> m_waveformContainerKey
    {this, "WaveformContainerKey", ""};
  //@}

};


#endif // SCINTDIGIALGS_SCINTDIGIALG_H
